import os
import zipfile
from abc import abstractmethod
from pathlib import Path
from typing import List, Optional, Union
from urllib.request import urlretrieve

import nltk

from .ressources import CACHE_PATH, CORENLP_URL
from .util import DownloadProgressBar


class CoreNLPParserInterface:
    # Path to the corenlp JAR models to use for parsing and create Tree
    # As of july 2023, Stanza does not return a Tree by a dictionary. Thus, we use NLTK API
    # that parse and return a dependency parse tree.
    CORENLP_DIRECTORY = "stanford-corenlp-full-2018-02-27"
    JAR_FILE_NAME = os.path.join(CORENLP_DIRECTORY, "stanford-corenlp-3.9.1.jar")
    JAR_MODEL_FILE_NAME = os.path.join(CORENLP_DIRECTORY, "stanford-corenlp-3.9.1-models.jar")

    def __init__(self, verbose: bool = True, cache_path: Optional[str] = None) -> None:
        """
         Create a constituency parsing model that use CoreNLP constituency parser. To do so, we download the latest
        model from CoreNLP (i.e. 2018) as suggest by this Wiki
        https://github.com/nltk/nltk/wiki/Stanford-CoreNLP-API-in-NLTK.

        :param verbose: (bool) Either or not to be verbose during the download of CoreNLP model. Default to `True`.
        :param cache_path: (Optional[str]) Optional parameter to set a cache path to download the CoreNP model to.
            If the cache_path is not set, the model are downloaded in the default cache path i.e. `'.cache/aaj'`.
        """

        if cache_path is None:
            cache_path = CACHE_PATH

        self.jar_file_name = os.path.join(cache_path, self.JAR_FILE_NAME)
        self.jar_model_file_name = os.path.join(cache_path, self.JAR_MODEL_FILE_NAME)

        self.verbose = verbose
        if not os.path.exists(self.jar_file_name) and not os.path.exists(self.jar_model_file_name):
            if self.verbose:
                reporthook = DownloadProgressBar()
            else:
                reporthook = None

            # Download zipped file with verbose report
            local_filename, _ = urlretrieve(CORENLP_URL, reporthook=reporthook)

            # Create .cache directory if it does not exist
            Path(cache_path).mkdir(parents=True, exist_ok=True)

            # Unzip the file into the cache directory
            with zipfile.ZipFile(local_filename, "r") as f:
                f.extractall(cache_path)

    @abstractmethod
    def tree_parser_sentences(self, sentences: List[str]) -> List[List[Union[str, nltk.tree.tree.Tree]]]:
        """
        Method to parse sentences into a tree.

        :param sentences: (list) A list of sentence to parse into trees.
        :return: A list of parse tree.
        """

    @property
    def tags_list(self) -> List[str]:
        tags_list = list(self._tags_mapping.keys())

        # We pop the first since it's the "__comment__", which is not a tag but the tags list source details.
        tags_list.pop(0)

        return tags_list



import json
import os
import pkgutil
from typing import List, Optional, Union

import conllu
import nltk
import supar
from nltk.parse.corenlp import CoreNLPServer, CoreNLPDependencyParser
from supar import Parser
from tqdm import tqdm

from .core_nlp_parser_interface import CoreNLPParserInterface


class DependencyParserCoreNLP(CoreNLPParserInterface):
    def __init__(self, verbose: bool = True, cache_path: Optional[str] = None) -> None:
        """
        Create a dependency parsing model that use CoreNLP dependency parser. To do so, we download the latest
        model from CoreNLP (i.e. 2018) as suggest by this Wiki
        https://github.com/nltk/nltk/wiki/Stanford-CoreNLP-API-in-NLTK.

        The tags are the Universal Dependencies tag V1 available here
        https://universaldependencies.org/docsv1/u/dep/index.html.

        :param verbose: (bool) Either or not to be verbose during the download of CoreNLP model. Default to `True`.
        :param cache_path: (Optional[str]) Optional parameter to set a cache path to download the CoreNP model to.
            If the cache_path is not set, the model are downloaded in the default cache path i.e. `'.cache/aaj'`.
        """
        super().__init__(verbose, cache_path)

        data = pkgutil.get_data(
            __name__,
            os.path.join(
                "./resources",
                "conll_u_tags_mapping.json",
            ),
        )
        self._tags_mapping = json.loads(data.decode("utf-8"))

    def tree_parser_sentences(
        self, sentences: List[str]
    ) -> List[List[Union[str, nltk.parse.dependencygraph.DependencyGraph]]]:
        """
        Method to parse sentences into dependency tree.

        :param sentences: (list) A list of sentence to parse into trees.
        :return: A list of NLTK dependency parse tree.
        """
        with CoreNLPServer(path_to_jar=self.jar_file_name, path_to_models_jar=self.jar_model_file_name) as server:
            parser = CoreNLPDependencyParser(url=server.url)
            parsed_trees = []
            if self.verbose:
                sentences = tqdm(sentences, total=len(sentences), desc="Processing dataset into trees")

            for sentence in sentences:
                if len(sentence) > 0:
                    # raw_parse return a list iterator, so we convert it into a list
                    # Once converted, it includes a dependency graph inside (seen as a defaultdict).
                    parsed_trees.append(list(parser.raw_parse(sentence)))
                else:
                    parsed_trees.append([""])
            return parsed_trees


class DependencyParserSuPar:
    def __init__(self, model: str, verbose: bool = True):
        """
        Create a dependency parsing model that use SuPar constituency parser.

        Base on the SuPar documentation https://github.com/yzhangcs/parser#usage.

        :param model: (str) The parsing model to use. Choices are
            - `'biaffine'` (https://openreview.net/forum?id=Hk95PK9le),
            - `'crf'` (https://aclanthology.org/2020.acl-main.302/),
            - `'crf2o'` (https://aclanthology.org/2020.acl-main.302/), and
            - `'vi'` (https://aclanthology.org/2020.aacl-main.12).
        :param verbose: (bool) Either or not to be verbose during the download of CoreNLP model. Default to `True`.
        """

        self.process_pipeline = Parser.load(f'{model}-dep-en')

        self.verbose = verbose

    def get_tree(self, sentence: supar.utils.Dataset) -> List[conllu.models.TokenTree]:
        """
        Interface method to get the tree depending on the sentence object.

        :param sentence: A SuPar Dataset.
        :return: Return a list of Tree SuPar Sentence.
        """
        return conllu.parse_tree(str(sentence.sentences[0]))

    def process_sentences(self, sentence: str) -> supar.utils.Dataset:
        """
        Interface method to process sentences.

        :param sentence: A sentence.
        :return: Return a SuPar dataset.
        """
        return self.process_pipeline.predict(sentence, lang="en", prob="False", verbose="False")

    def tree_parser_sentences(self, sentences: List[str]) -> List[List[Union[str, conllu.models.TokenTree]]]:
        """
        Method to parse sentences into constituency tree.

        :param sentences: (list) A list of sentence to parse into trees.
        :return: A list of SuPar parse tree.
        """
        parsed_trees = []

        if self.verbose:
            sentences = tqdm(sentences, total=len(sentences), desc="Processing dataset into trees")

        for sentence in sentences:
            if len(sentence) > 0:
                process_documents = self.process_sentences(sentence)
                parsed_trees.append(self.get_tree(process_documents))
            else:
                parsed_trees.append([""])
        return parsed_trees



import json
import os
import pkgutil
from typing import List, Tuple

import math
import numpy as np

from .fluency_score_interface import LanguageModelFluencyScoreInterface


class NCE(LanguageModelFluencyScoreInterface):
    def __init__(
        self,
        language_model: str,
        device: str = "cuda",
        tokenizer: str = "treebank",
        do_normalization: bool = True,
        rescale_bound_ab: Tuple = (0, 1),
    ):
        """
        Instantiate a Negative Cross-Entropy metric to compute the NCE score or the WPNCE.

        NCE is equal to the probability of occurrence of a given sentence base on the likelihood of a
        language model normalizes by the number of word in the sentence (i.e. sentence length).

        Args:
            language_model (str): The language model to compute the sentence and pre-process word probabilities matrix.
                The supported model are ``"bert-base-uncased"``, ``"bert-large-uncased"`` and ``"roberta-base"``.
            device (str): The device to use such as ``"cpu"``, ``"cuda"`` or ``"cuda:device_int"``.
            tokenizer(str): The tokenizer to use to split the sentence. If the tokenizer is set to ``"treebank"``,
                ``"whitespace"`` or ``"whitespace_lm"`` it will be a NCE metric. If the tokenizer is set to
                ``"wordpiece"``, it is a WPNCE. Other than the tokenizing process, the compute method is similar.
            do_normalization (bool): Either or not to do normalization. By default, set to True.
            rescale_bound_ab (Tuple): A tuple to specify the rescaling normalization bound, by default the range is
                [0, 1].

        Implementation details:

            See LanguageModelFluencyScoreInterface for implementation details.
        """

        super().__init__(language_model, device, tokenizer, do_normalization, rescale_bound_ab)

        files_name = f"{tokenizer.lower()}"
        if tokenizer.lower() in ("whitespace_llm", "wordpiece"):
            files_name += f"_{language_model.lower().replace('-', '_')}"

        data = pkgutil.get_data(
            __name__,
            os.path.join(
                "./resources",
                "nce_scores_distribution",
                f"{files_name}.json",
            ),
        )

        # Computed on the Gigaword dataset version 1.2.0
        # The min score is mu - 2 * sigma (i.e. the mean minus 2 standard deviation)
        # The max score is mu + 2 * sigma (i.e. the mean plus 2 standard deviation)
        # Statistically, 95.45% of the scores are include in this range.
        # https://en.wikipedia.org/wiki/68%E2%80%9395%E2%80%9399.7_rule
        nce_scores_distribution = json.loads(data.decode("utf-8"))
        mu = nce_scores_distribution.get("mean")
        standard_deviation = math.sqrt(nce_scores_distribution.get("variance"))
        self.normalisation_min_score = mu - 2 * standard_deviation
        self.normalisation_max_score = mu + 2 * standard_deviation

    def compute_score(self, sentences: List[str]) -> List[float]:
        """
        Computed base of this article https://arxiv.org/pdf/1809.08731v1.pdf.
        """

        (
            padded_sentences_words_idx,
            sentences_mask,
            batch_logits,
        ) = self._llm_batch_processing(sentences=sentences)

        sentences_len: List[int] = self._compute_sentences_len(sentences_mask)

        sentences_log_prob = self._compute_pm(padded_sentences_words_idx, sentences_mask, batch_logits, sentences_len)

        sentences_log_prob = np.nan_to_num(sentences_log_prob)

        nce_score: List = list(np.divide(sentences_log_prob, sentences_len))

        if self.do_normalization:
            nce_score = self._normalize_scores(
                scores=nce_score,
                x_min=self.normalisation_min_score,
                x_max=self.normalisation_max_score,
                a=self.normalisation_a_score,
                b=self.normalisation_b_score,
            )
        return nce_score
